﻿-- ********************************************************************************
-- Broker Statistics
-- A Data Broker display for combat statistics.
-- By: Shenton
--
-- Core.lua
-- Ace3 libs, core functions, Ace3 init.
-- ********************************************************************************
--########################################################################################################################################################################################
-- TODO laisser le joueur choisr si il veut recharger l'interface ou non après avoir désactivé un objet ldb
--########################################################################################################################################################################################
-- Ace libs (<3)
local A = LibStub("AceAddon-3.0"):NewAddon("BrokerStats", "AceConsole-3.0", "AceTimer-3.0", "AceComm-3.0", "AceEvent-3.0");
A.L = LibStub("AceLocale-3.0"):GetLocale("BrokerStats");
A.AceConfigDialog = LibStub("AceConfigDialog-3.0");
A.AceDBOptions = LibStub("AceDBOptions-3.0");
A.AceConfigRegistry = LibStub("AceConfigRegistry-3.0");

-- LibDataBroker
A.LibDataBroker = LibStub("LibDataBroker-1.1");

-- Addon global
_G["DataBrokerStatisticsGlobal"] = A;

-- LUA globals
local mfloor, pairs, ipairs, loadstring = math.floor, pairs, ipairs, loadstring;
local smatch, sgsub, tostring, type = string.match, string.gsub, tostring, type;
local tinsert, tconcat, format, tsort = table.insert, table.concat, string.format, table.sort;
local ssub = string.sub;

-- WoW globals
local strtrim = strtrim;

-- ********************************************************************************
-- Functions
-- ********************************************************************************

--- Send a message to the chat frame with the addon name colored
-- @param text The message to display
-- @param level Verbosity level (0 to 2)
-- @param color Bool, if true will color in red
function A:Message(text, level, color)
    if ( level <= A.db.profile.verbosity ) then
        if ( color ) then
            color = A.colorRed;
        else
            color = A.colorGreen;
        end

        DEFAULT_CHAT_FRAME:AddMessage(color..A.L["Broker Statistics"]..": "..A.colorReset..text);
    end
end

--- Handle the slash command
function A:SlashCommand()
    A.AceConfigDialog:Open("BrokerStatsConfig");
end

--- Round function (http://lua-users.org/wiki/SimpleRound)
-- @param num The number to round
-- @param idp Decimal places
function A:Round(num, idp)
    if ( type(num) == "number" ) then
        local mult = 10^(idp or 0);
        return mfloor(num * mult + 0.5) / mult;
    else
        A:Message(A.L["Attempt to perform arithmetic on a non number value, you are probably trying to round a string."], 1, 1);
        return num;
    end
end

--- Color percent to hex
local b16 =
{
    [0] = "0",
    [1] = "1",
    [2] = "2",
    [3] = "3",
    [4] = "4",
    [5] = "5",
    [6] = "6",
    [7] = "7",
    [8] = "8",
    [9] = "9",
    [10] = "a",
    [11] = "b",
    [12] = "c",
    [13] = "d",
    [14] = "e",
    [15] = "f"
};
function A:PercentToHex(r, g, b, a)
    r = A:Round(r * 255);
    g = A:Round(g * 255);
    b = A:Round(b * 255);
    a = A:Round(a * 255);

    r = b16[mfloor(r / 16)]..b16[r % 16];
    g = b16[mfloor(g / 16)]..b16[g % 16];
    b = b16[mfloor(b / 16)]..b16[b % 16];
    a = b16[mfloor(a / 16)]..b16[a % 16];

    return a..r..g..b;
end

--- Convert table to string (http://lua-users.org/wiki/TableUtils)
function A:TableValToStr(v)
    if ( "string" == type(v) ) then
        v = sgsub(v, "\n", "\\n");
        if ( smatch(sgsub(v, "[^'\"]", ""), '^"+$') ) then return "'"..v.."'"; end
        return '"'..sgsub(v, '"', '\\"')..'"';
    else
        return "table" == type(v) and A:TableTostring(v) or tostring(v);
    end
end
function A:TableKeyToStr(k)
    if ( "string" == type(k) and smatch(k, "^[_%a][_%a%d]*$") ) then return k; end

    return "["..A:TableValToStr( k ) .. "]";
end
function A:TableTostring(tbl)
    local result, done = {}, {};

    for k,v in ipairs(tbl) do
        tinsert(result, A:TableValToStr(v));
        done[k] = true;
    end

    for k, v in pairs(tbl) do
        if ( not done[k] ) then tinsert(result, A:TableKeyToStr(k).."="..A:TableValToStr(v)); end
    end

    return "{"..tconcat(result, ",").."}";
end

--- Pairs function with alphabetic sort
function A:PairsByKeys(t, f)
    local a, i = {}, 0;

    for n in pairs(t) do a[#a+1] = n; end
    tsort(a, f);

    local iter = function()
        i = i + 1;
        if ( not a[i] ) then
            return nil;
        else
            return a[i], t[a[i]];
        end
    end

    return iter;
end

--- Remove unused entries from the database
function A:DelUnusedEntry()
    for k in pairs(A.db.profile.buildInStats) do
        for kk,vv in pairs(A.db.profile.buildInStats[k]) do
            if ( vv.dbObject ) then
                local ok;

                for kkk,vvv in pairs(A.db.global.customDisplays) do
                    if ( vvv ) then ok = 1; end
                end

                if ( not ok ) then
                    A.db.profile.buildInStats[k][kk].dbObject = "Default";
                end
            end
        end
    end

    for k in pairs(A.db.profile.custom) do
        for kk,vv in pairs(A.db.profile.custom[k]) do
            if ( vv.dbObject ) then
                local ok;

                for kkk,vvv in pairs(A.db.global.customDisplays) do
                    if ( vvv ) then ok = 1; end
                end

                if ( not ok ) then
                    A.db.profile.custom[k][kk].dbObject = "Default";
                end
            end
        end
    end
end

-- Return if there is a least one custom stat
function A:GotCustom()
    for _,v in pairs(A.db.global.custom) do
        if ( v.calc and v.calc ~= "" ) then return 1; end
    end

    return nil;
end

-- Return if there is a least one custom ldb object
function A:GotCustomLdb()
    for k in pairs(A.db.global.customDisplays) do
        if ( k ) then return 1; end
    end

    return nil;
end

--- Create the select table for ldb objects
local out;
function A:CreateDbObjectsSelect()
    out =
    {
        [A.L["Default"]] = A.L["Default"];
    };

    for k in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        out[k] = k;
    end

    return out;
end

--- Return informations about the current specialisation
function A:GetSpecInfos()
    local specGroup = GetActiveTalentGroup(false, false);
    local numTabs = GetNumTalentTabs(false, false);
    local specName, nameOut, last, point, numOut = nil, nil, 0, nil, "";

    for i=1,numTabs do
        _, specName,_,_, points = GetTalentTabInfo(i, false, false, specGroup)

        if ( points > last ) then nameOut = specName; end

        numOut = numOut.."/"..points;
        last = points;
    end

    numOut = "("..strtrim(numOut, "/")..")";

    return nameOut, numOut;
end

--- Set the current spec var to the player current spec
function A:SetSpecVar()
    if ( A.db.profile.ConfigAccordingSpec ) then
        local name, num = A:GetSpecInfos();

        if ( A.db.profile.ConfigAccordingSpecBehavior == 2 ) then
            A.playerSpec = name.." "..num;
        else
            A.playerSpec = name;
        end
    else
        A.playerSpec = "Default";
    end
end

--- Pause toggle
function A:PauseToggle()
    if ( A.paused ) then
        A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
        A.paused = nil;
    else
        A:CancelTimer(A.mainTimer, 1);
        A:SetTextOnLdbObjects(A.colorRed..A.L["Paused"]);
        A.paused = 1;
    end
end

--- Refresh the main timer to the current delay
function A:RefreshTimer()
    A.db.profile.time = A:Round(A.db.profile.time, 1);
    A:CancelTimer(A.mainTimer, 1);
    A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
end

--- Count the number of activated statistics and ldb objects
-- @return the number of stats and objects activated and their names
function A:CountActivated()
    local stats, countStats, objects, countObjects = {}, 0, {}, 0;

    for k,v in pairs(A.db.profile.buildInStats[A.playerSpec]) do
        if ( v.enabled ) then
            stats[#stats+1] = k;
            countStats = countStats + 1;
        end
    end

    for k,v in pairs(A.db.profile.custom[A.playerSpec]) do
        if ( v.enabled ) then
            stats[#stats+1] = k;
            countStats = countStats + 1;
        end
    end

    for k,v in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( v.enabled ) then
            objects[#objects+1] = k;
            countObjects = countObjects + 1;
        end
    end

    return stats, countStats, objects, countObjects;
end

--- Display a string on all enabled ldb objects
function A:SetTextOnLdbObjects(text)
    A.ldbObjectDefault.text = text;

    for k,v in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( v.enabled ) then A["ldbObject"..k].text = text; end
    end
end

--- Create the others Data Broker objects
function A:CreateDataBrokerOjbects()
    for k,v in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( v.enabled ) then
            local icon;

            if ( v.icon ) then
                icon = v.icon;
            else
                icon = A.defaultLdbIcon;
            end

            A:CreateDataBrokerOjbect(k, icon);
        end
    end
end

--- Create a Data Broker objects if needed
function A:CreateDataBrokerOjbect(name, icon)
    if ( not name ) then name = "Default"; end

    if ( A.LibDataBroker:GetDataObjectByName("Broker Stats - "..name) ) then
        A["ldbObject"..name] = A.LibDataBroker:GetDataObjectByName("Broker Stats - "..name);
        A["ldbObject"..name].text = A.L["Nothing to display"];
        A["ldbObject"..name].label = A.L["Broker Statistics"];
        A["ldbObject"..name].icon = icon;
        A["ldbObject"..name].OnTooltipShow = function(tooltip)
            local statsNames, countStats, objectsName, countObjects = A:CountActivated();

            objectsName[#objectsName+1] = A.L["Default"];

            statsNames = tconcat(statsNames, ", ");
            objectsName = tconcat(objectsName, ", ");

            tooltip:AddDoubleLine(A.colorWhite..A.L["Broker Statistics"], A.colorGreen.."v"..A.version);
            tooltip:AddLine(" ");
            tooltip:AddLine(A.L["Statistics displayed: %d"]:format(countStats));
            tooltip:AddLine(statsNames);
            tooltip:AddLine(" ");
            tooltip:AddLine(A.L["Data broker objects used: %d"]:format(countObjects + 1));
            tooltip:AddLine(objectsName);
            tooltip:AddLine(" ");
            tooltip:AddLine(A.L["|cFFC79C6ELeft-Click: |cFF33FF99Pause the addon update timer\n|cFFC79C6ERight-Click: |cFF33FF99Open the quick access menu\n|cFFC79C6EMiddle-Click: |cFF33FF99Open the configuration panel"]);
        end
        A["ldbObject"..name].OnClick = function(self, button)
            if (button == "LeftButton") then
                A:PauseToggle();
            elseif ( button == "RightButton" ) then
                A:DropDownMenu(self, "Default");
            elseif ( button == "MiddleButton" ) then
                A.AceConfigDialog:Open("BrokerStatsConfig");
            end
        end
    else
        A["ldbObject"..name] = A.LibDataBroker:NewDataObject("Broker Stats - "..name,
        {
            type = "data source",
            text = A.L["Nothing to display"],
            label = A.L["Broker Statistics"],
            icon = icon,
            tocname = "Broker_Stats",
            OnClick = function(self, button)
                if (button == "LeftButton") then
                    A:PauseToggle();
                elseif ( button == "RightButton" ) then
                    A:DropDownMenu(self, "Default");
                elseif ( button == "MiddleButton" ) then
                    A.AceConfigDialog:Open("BrokerStatsConfig");
                end
            end,
            OnTooltipShow = function(tooltip)
                local statsNames, countStats, objectsName, countObjects = A:CountActivated();

                objectsName[#objectsName+1] = A.L["Default"];

                statsNames = tconcat(statsNames, ", ");
                objectsName = tconcat(objectsName, ", ");

                tooltip:AddDoubleLine(A.colorWhite..A.L["Broker Statistics"], A.colorGreen.."v"..A.version);
                tooltip:AddLine(" ");
                tooltip:AddLine(A.L["Statistics displayed: %d"]:format(countStats));
                tooltip:AddLine(statsNames);
                tooltip:AddLine(" ");
                tooltip:AddLine(A.L["Data broker objects used: %d"]:format(countObjects + 1));
                tooltip:AddLine(objectsName);
                tooltip:AddLine(" ");
                tooltip:AddLine(A.L["|cFFC79C6ELeft-Click: |cFF33FF99Pause the addon update timer\n|cFFC79C6ERight-Click: |cFF33FF99Open the quick access menu\n|cFFC79C6EMiddle-Click: |cFF33FF99Open the configuration panel"]);
            end
        });
    end
end

--- Hide disabled ldb objects
function A:HideDisabledLdb()
    for k,v in pairs(A.db.global.customDisplays) do
        if ( not A.db.profile.dataBroker.customDisplays[A.playerSpec][k] ) then
            A["ldbObject"..k].text = "";
            A["ldbObject"..k].label = "";
            A["ldbObject"..k].icon = "";
            A["ldbObject"..k].OnTooltipShow = nil;
            A["ldbObject"..k].Onclick = nil;
        elseif ( not A.db.profile.dataBroker.customDisplays[A.playerSpec][k].enabled ) then
            A["ldbObject"..k].text = "";
            A["ldbObject"..k].label = "";
            A["ldbObject"..k].icon = "";
            A["ldbObject"..k].OnTooltipShow = nil;
            A["ldbObject"..k].Onclick = nil;
        end
    end
end

--- Create the needed (empty) tables within the config db if needed
function A:CreateNeededConfigTables()
    if ( not A.db.profile.dataBroker.customDisplays[A.playerSpec] ) then A.db.profile.dataBroker.customDisplays[A.playerSpec] = {}; end
    if ( not A.db.profile.buildInStats[A.playerSpec] ) then A.db.profile.buildInStats[A.playerSpec] = {}; end
    if ( not A.db.profile.custom[A.playerSpec] ) then A.db.profile.custom[A.playerSpec] = {}; end
end

--- Create statistics table
local customDisplays;
function A:CreateStatisticsTable()
    customDisplays =
    {
        Default = 1
    };

    A.statsTable = {};

    for k,v in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( v.enabled ) then customDisplays[k] = 1; end
    end

    for k,v in pairs(A.db.profile.buildInStats[A.playerSpec]) do
        if ( v.enabled ) then
            local opt = A.buildInStats[k];
            local dbObject = v.dbObject;

            if ( not dbObject ) then dbObject = "Default"; end

            if ( customDisplays[dbObject] ) then
                A.statsTable[k] = {};

                if ( opt.pre and A.db.profile.displayPre ) then
                    A.statsTable[k].pre = A.db.profile.preColorHex..opt.pre..A.colorReset;
                else
                    A.statsTable[k].pre = "";
                end
                if ( opt.post and A.db.profile.displayPost ) then
                    A.statsTable[k].post = A.db.profile.postColorHex..opt.post..A.colorReset;
                else
                    A.statsTable[k].post = "";
                end
                if ( v.iconEnabled ) then
                    if ( v.icon ) then
                        A.statsTable[k].icon = "|T"..v.icon..":"..A.db.profile.statIconSize.."|t";
                    else
                        A.statsTable[k].icon = "|T"..A.defaultStatIcon..":"..A.db.profile.statIconSize.."|t";
                    end
                else
                    A.statsTable[k].icon = "";
                end
                if ( v.colorHex ) then
                    A.statsTable[k].color = v.colorHex;
                else
                    A.statsTable[k].color = "";
                end
                A.statsTable[k].dbObject = dbObject;
                A.statsTable[k].multiple = opt.multiple;
                A.statsTable[k].round = opt.round;
                A.statsTable[k].separator = opt.separator;
                A.statsTable[k].calc = loadstring(opt.calc);
            end
        end
    end

    for k,v in pairs(A.db.profile.custom[A.playerSpec]) do
        if ( v.enabled ) then
            local opt = A.db.global.custom[k];
            local calc, dbObject = nil, v.dbObject;

            if ( not dbObject ) then dbObject = "Default"; end

            if ( customDisplays[dbObject] ) then
                if ( opt.calc ) then calc = loadstring(opt.calc); end

                local ok, errorString = pcall(calc);

                if ( ok ) then
                    if ( calc ) then
                        A.statsTable[k] = {};

                        if ( opt.pre and A.db.profile.displayPre ) then
                            A.statsTable[k].pre = A.db.profile.preColorHex..opt.pre..A.colorReset;
                        else
                            A.statsTable[k].pre = "";
                        end
                        if ( opt.post and A.db.profile.displayPost ) then
                            A.statsTable[k].post = A.db.profile.postColorHex..opt.post..A.colorReset;
                        else
                            A.statsTable[k].post = "";
                        end
                        if ( v.iconEnabled ) then
                            if ( v.icon ) then
                                A.statsTable[k].icon = "|T"..v.icon..":"..A.db.profile.statIconSize.."|t";
                            else
                                A.statsTable[k].icon = "|T"..A.defaultStatIcon..":"..A.db.profile.statIconSize.."|t";
                            end
                        else
                            A.statsTable[k].icon = "";
                        end
                        if ( v.colorHex ) then
                            A.statsTable[k].color = v.colorHex;
                        else
                            A.statsTable[k].color = "";
                        end
                        A.statsTable[k].dbObject = dbObject;
                        A.statsTable[k].multiple = opt.multiple;
                        A.statsTable[k].round = opt.round;
                        A.statsTable[k].separator = opt.separator;
                        A.statsTable[k].calc = calc;
                    else
                        A:Message(A.L["There is an error in %s code. Error: %s"]:format(k, A.L["Returned value is nil."]), 1, 1);
                    end
                else
                    A:Message(A.L["There is an error in %s code. Error: %s"]:format(k, errorString), 1, 1);
                end
            end
        end
    end
end

--- Update text on the data broker display
local updateText;
function A:UpdateText()
    updateText = {};

    for k,v in pairs(A.statsTable) do
        local stat = "";

        if ( v.multiple ) then
            local values = { v.calc() };

            for _,vv in pairs(values) do
                if ( v.round ) then
                    stat = stat..A:Round(vv, v.round)..v.post..v.separator;
                else
                    stat = stat..vv..v.post..v.separator;
                end
            end

            stat = strtrim(stat, v.separator);

            if ( not updateText[v.dbObject] ) then
                updateText[v.dbObject] = v.pre..v.icon..v.color..stat..A.colorReset;
            else
                updateText[v.dbObject] = updateText[v.dbObject].." "..A.db.profile.separatorColorHex..A.db.profile.statsSeparator.." "..A.colorReset..
                v.pre..v.icon..v.color..stat..A.colorReset;
            end
        else
            if ( v.round ) then
                stat = A:Round(v.calc(), v.round);
            else
                stat = v.calc();
            end

            if ( not updateText[v.dbObject] ) then
                updateText[v.dbObject] = v.pre..v.icon..v.color..stat..A.colorReset..v.post;
            else
                updateText[v.dbObject] = updateText[v.dbObject].." "..A.db.profile.separatorColorHex..A.db.profile.statsSeparator.." "..A.colorReset..
                v.pre..v.icon..v.color..stat..A.colorReset..v.post;
            end
        end
    end

    for k,v in pairs(updateText) do
        if ( v == "" ) then v = A.L["Nothing to display"]; end

        A["ldbObject"..k].text = v;
    end
end

-- ********************************************************************************
-- Addon Communication
-- ********************************************************************************

--- Called when receiving an addon communication message
function A:AddonComm(...)
    local prefix, customText, method, who, customName;

    prefix, customText, method, who = ...;

    if ( prefix ~= "BrokerStatsComm" ) then return; end
    if ( who == UnitName("player") ) then return; end

    if ( (method == "WHISPER" and A.db.profile.shareWhisper) or
    (method == "RAID" and A.db.profile.shareRaid) or
    (method == "GUILD" and A.db.profile.shareGuild) or
    (method == "BATTLEGROUND" and A.db.profile.shareBg) ) then
        A:Message(A.L["%s tried to send you a custom stat (%s) with method %s."]:format(who, customName, method), 2, 1);
        return;
    end

    customName = smatch(customText, "^(.+){");
    customText = smatch(customText, "{(.+)}");
    customText = "{"..customText.."}";

    if ( A.db.global.custom[customName] ) then
        local pop = StaticPopup_Show("BrokerStatsReceiveCustomNameExists", customName, who);
        if (pop) then pop.customText = customText; end
    elseif ( A.buildInStats[customName] ) then
        local pop = StaticPopup_Show("BrokerStatsReceiveCustomNameExists", customName, who);
        if (pop) then pop.customText = customText; end
    else
        local pop = StaticPopup_Show("BrokerStatsReceiveCustom", customName, who);
        if (pop) then
            pop.customText = customText;
            pop.customName = customName;
        end
    end
end

--- Send an addon communication message with a custom stat
-- @param target The message target
-- @param customStat The custom stat name
function A:SendStat(target, customStat)
    if ( not target or not customStat ) then return; end

    target = strtrim(target);
    if ( target == "" ) then return; end

    local textMessage = A:TableTostring(A.db.global.custom[customStat]);

    if ( target == "RAID" or target == "GUILD" or target == "BATTLEGROUND" ) then
        A:SendCommMessage("BrokerStatsComm", customStat..textMessage, target);
    else
        A:SendCommMessage("BrokerStatsComm", customStat..textMessage, "WHISPER", target);
    end
end

-- ********************************************************************************
-- Callbacks
-- ********************************************************************************

--- Called when an action is made on a configuration profile
function A:RefreshConfig()
    A:SetSpecVar();
    A:CreateNeededConfigTables();
    A.AceConfigRegistry:NotifyChange("BrokerStatsConfig");
    A:RefreshTimer();
    A:CreateDataBrokerOjbects();
    A:HideDisabledLdb()
    A:CreateStatisticsTable();
end

--- Callback for event ACTIVE_TALENT_GROUP_CHANGED
function A:ACTIVE_TALENT_GROUP_CHANGED()
    A:SetSpecVar();
    A:CreateNeededConfigTables();
    A.AceConfigRegistry:NotifyChange("BrokerStatsConfig");
    A:RefreshTimer();
    A:CreateDataBrokerOjbects();
    A:HideDisabledLdb();
    A:CreateStatisticsTable();
end

-- ********************************************************************************
-- Main
-- ********************************************************************************

--- AceAddon callback
-- Called after the addon is fully loaded
function A:OnInitialize()
    -- Database
    A.db = LibStub("AceDB-3.0"):New("BrokerStatsDB", A:AceDefaultDB());

    -- Config settings refresh
    A.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig");
    A.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig");
    A.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig");

    -- Slash command
    A:RegisterChatCommand("brokerstats", "SlashCommand");
    A:RegisterChatCommand("stats", "SlashCommand");
    A:RegisterChatCommand("bs", "SlashCommand");
end

--- AceAddon callback
-- Called during the PLAYER_LOGIN event
function A:OnEnable()
    -- Config
    LibStub("AceConfig-3.0"):RegisterOptionsTable("BrokerStatsConfig", A.AceConfig);
    A.AceConfigDialog:SetDefaultSize("BrokerStatsConfig", 800, 500);
    A.AceConfigDialog:AddToBlizOptions("BrokerStatsConfig", A.L["Broker Statistics"]);

    -- Default LDB object
    A:CreateDataBrokerOjbect(nil, A.db.profile.dataBroker.defaultObjectIcon);

    -- DropDownMenu frame
    A.dropDownMenuFrame = CreateFrame("Frame", "BrokerStatisticsMenuFrame");
    A.dropDownMenuFrame.displayMode = "MENU";

    -- Create "Default" config tables
    A:CreateNeededConfigTables();

    -- Delete unused database entries
    A:DelUnusedEntry();

    -- Events
    A:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");

    -- Register for addon communication
    A:RegisterComm("BrokerStatsComm", "AddonComm");

    -- Player spec
    A:SetSpecVar();

    -- Create Custom ldb objects
    A:CreateDataBrokerOjbects();

    -- Create the statistics table
    A:CreateStatisticsTable();

    -- Main timer
    A.db.profile.time = A:Round(A.db.profile.time, 1);
    A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
end

--- AceAddon callback
-- Called when the addon is manually disabled
function A:OnDisable()
    A:CancelTimer(A.mainTimer, 1);
    A:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
end
